/*
  gmMakeMasks_2.cc

  AUTHOR: Dustin Fennell - University of Victoria - Feb 2010
  Based partially on gmMakeMasks.cc by B. Ottini, Craig Allen, Jennifer Dunn
  DATE CREATED:
  DATE UPDATED: August 18, 2011, Bryan Miller

  PURPOSE: To fit as many slits as possible onto a GMOS or F2 mask prototype.
  ( and write the data from those slits into an ODF file(s) )

  OVERVIEW:
  gmMakeMasks_2.cc is part of the larger GMMPS software, and is
  invoked from gmmps_spoc.tcl.

  The program reads data from a temporary data file created by GMMPS, and
  outputs at least one ODF file in a format that is understood by the rest
  of the GMMPS program.

  gmMakeMasks_2.cc is loosely based on gmMakeMasks.cc, and provides most of
  the functionality of that file. It is lacking GMOS nod-and-shuffle and
  microshuffle support, and as such gmMakeMasks.cc is still maintained and
  invoked within the GMMPS system.

  ALGORITHM SUMMARY:
  The algorithm chosen for slit selection is reasonably simple, but seems to
  be also fairly effective (especially if compared to the previous algorithm
  used in gmMakeMasks).

  Slits on the mask are first represented as a graph, with each slit as a
  vertex. Edges are then placed where two slits cannot be placed on the
  same graph (ie when their spectra, or the slits themselves, overlap). After
  the slits are represented this way the slit selection process begins.
  For each priority (In order: Acquisition, Priority 1, 2, and 3) the slit
  with the least edges is placed on the graph. Then all slits that were
  adjacent to the placed slit are removed from the graph, and the graph
  is updated (edges to slits no longer on the graph are cleared). This
  process is repeated until all slits in the priority level have been placed
  or excluded from placement, at which point the next priority level is
  processed.

  POSSIBLE FUTURE OPTIMIZATIONS:
  There are several places where this program can be optimized, both in
  run-time efficiency and in terms of the number of slits placed by the
  algorithm.
  As of this moment (08-Mar-2010) the program runs in polynomial time. It
  is able to do its computations for reasonably sized data sets very quickly,
  therefor code optimizations are not in my opinion required at this
  point in time.

  Improving the slit-placement algorithm itself is always welcome, however..
  especially in the situation where we have clock-cycles to spare.
  I had a few ideas as to how this code might be improved, which I will
  outline below:

  1.) Brute-force slit selection for small data sets.
  - If the set of slits to be placed/excluded is below a certain threshold
  then a brute-force algorithm (looking at all possible slit placement
  combinations and picking the one with most slits placed) could improve
  things a bit. I was thinking that a threshold of 10-20 slits would be
  optimal, though once the algorithm is in place perhaps the appropriate
  number could be found by experiment.

  NOTE: I've implemented a version of 'brute force' below... it doesn't
  seem to work that well.

  2.) Same-degree slit selection improvement.
  - Currently (08-Mar-2010) when more than one slit of the same degree is
  found in a graph gmMakeMasks_2.cc places the first slit of this degree
  encountered (as things stand this will be the slit with the lowest id).
  This algorithm can perhaps be improved... perhaps looking for slits
  within this same-degree set that are themselves connected, and selecting
  the slit with the lowest degree from within those (though this may not be
  of that large a benefit).

  3.) Finding and exploiting sub-graphs.
  - It is very likely (especially in lower priorities) that the generated
  graph will not be a connected graph. As such it will consist of a number
  of sub-graphs.
  If the algorithm were smart enough to separate the slits into their
  respective sub-graphs, it could preform more efficient operations on those
  graphs (for example, if a set of 100 slits is in fact 20 sets of 5 slits,
  then a brute-force approach would be feasible and optimal).

  4.) Cut-vertices. (or similar effects)
  - By the nature of the problem, each time a slit is placed a cut will be
  made in the graph. If we were to select slits such that efficient cuts
  were made to the graph (using, perhaps, a combination of point 2 and 3,
  above) that could lead to efficiency improvements in both time and
  slit selection. We should exercise some caution here, as we wouldn't want
  to concentrate only on making good cuts at the expense of selecting more
  slits, but, if all things were equal we might as well make a better cut,
  than worse (probably).


  There are, I'm sure, many more optimizations that could be made to this code.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>
#include <string>
#include <vector>
#include <map>
#include <utility>
#include <iostream>
#include <fstream>
#include <sstream>
using namespace std;

// The minimum distance between two spectra (in arcsec, 4 old unbinned GMOS pixels)
float MIN_SPEC_DIST = 0.300;

// Minimum distance between two slits for GMOS (in Arcseconds)
#define MIN_GMOS_CUT 0.14  // unused???

/*
 * ---------------- Helper Class Prototypes ----------------
 */

class Slit {
public:
  int id;
  float slitStart, slitEnd, slitTop, slitBottom, ccdL, ccdW, slitLength,
    slitWidth, specPosL, specPosW, angle, wiggleRoom, maxWiggle;
  char priority;
  string line;
  bool wiggleUsed;

  Slit();
  Slit(int, char);
  Slit(int, char, float, float);
  Slit(int, char, float, float, float, float, float, float, float, float,
       float, float, float, float, string);
  int changePosition(float, float, float, string);
  bool stringToFloat(const string&, float&);
  vector<string> stringSplit(string, string);
};

class GraphNode {
public:
  int id, degree;
  vector<int> adjList;

  GraphNode();
  GraphNode(int);
  void addAdj(int);
  void removeAdj(int);
  vector<int> getList();
  int getDegree();
};

class Graph {
public:
  map<int, GraphNode> elements;

  void addNode(int);
  void removeNode(int);
  void clear();
  void addEdge(GraphNode*, GraphNode*);
  void removeEdge(GraphNode*, GraphNode*);
  GraphNode * getNode(int nid);
  void clearEdges(GraphNode * node);
  bool empty();
  int size();
  vector<int> getMinDegree();
};

/**
 * ------------------- Slit Selection Function Prototypes -----------------------
 **/

void initOutputFile(char*, ofstream&, char*, float, string, float, char*,
		    char*, char*, char*, float);
float loadSlits(map<int, Slit>*, char*, char*, char*, float, float, float,
		float, float, float, string, float, vector<float>*, float*, float*);
map<int, Slit> getSlits(map<int, Slit>, char);
void mapConflicts(Graph*, map<int, Slit>, float, float, string, float);
void placeSlits(map<int, Slit>*, map<int, Slit>*, map<int, Slit>*,
		map<int, Slit>*, Graph*, float, float, string, float, char*);
void removeConflicts(map<int, Slit>*, map<int, Slit>*, map<int, Slit>*, Graph);
vector<float> loadFov(char*, float, string);
bool bandShuffleCheck(vector<float>*, float, float, float);
void removeSlit(int, map<int, Slit>*, map<int, Slit>*);
void writeSlits(map<int, Slit>&, ofstream&);
void maxSlitMode(map<int, Slit>&, float, string, char*);
void maxModeEnforceBands(map<int, Slit>&, int, vector<float>&, string, float);
void maxOptProcessLine(map<int, Slit>&, int, string, float);
void expandSlitToFOV(map<int, Slit>&, vector<float>&, int, char);
bool conflicts(Slit, Slit, float, float, string, float);
int findMaxSpectrumSid(vector<int>*, map<int, Slit>*, float);
bool wiggleNear(int, map<int, Slit>*, map<int, Slit>*, float, string);
int wiggleUnplaced(map<int, Slit>*, map<int, Slit>*, map<int, Slit>*, float,
		   float, string, float);
bool checkConflicts(Slit, Slit, map<int, Slit>, float, float, string, float);
void printIntro(char, string);
void printIntroError(int);
vector<string> stringSplit(string, string);
bool stringToInt(const string&, int&);
bool stringToFloat(const string&, float&);

/**
 * ------------------- Slit Selection Code -----------------------
 **/

/*
************************************************************************
*+
* FUNCTION: main
*
* RETURNS: int [success or failure]
*
* DESCRIPTION: Main entry point to function.
*
* [NOTES:]:
*-
************************************************************************
*/
int main(int argc, char *argv[]) {
  // Input variables (set from argv).
  char inFile[256];      // Input file
  char outFile[256];     // Output file
  char outFileRoot[256]; // Output root path
  char fovFile[256];     // Field-of-View file
  string instType;
  char slitMode;         // Determines whether to expand slits into empty space or not.
  float specLength;      // spectrum length (pixels)
  float anamorphic;      // anamorphic factor
  float dPix;            // linear dispersion (nanometers/pixel)
  float pixelScale;      // pixel scale    (arcseconds/pixel)
  float l1Max;           // "lambda 1 max"
  float l2Min;           // "lambda 2 min"
  float minpixsep = 4.0; // the minimum separation of spectra in pixels for auto expansion
  int centralWavelength; // central wavelength (nanometers)
  int numMasks;          // number of masks to be made
  char det_img[256];
  char det_spec[256];
  char ra_img[256];
  char dec_img[256];
  float wiggleVal;
  char bdfile[256];
  bdfile[0] = 0;
  
  // Slit containers.
  map<int, Slit> slits;    // holds all slits
  map<int, Slit> acqSlits; // holds acquisition objects
  map<int, Slit> p1Slits;  // holds p1 objects
  map<int, Slit> p2Slits;  // holds p2 objects
  map<int, Slit> p3Slits;  // holds p3 objects
  map<int, Slit> placed;   // holds objects which to be placed in the ODF
  map<int, Slit> removed;  // holds objects not in ODF
  
  // Holds band boundaries in nod and shuffle mode (band shuffle).
  vector<float> bandLocations;
  float shuffleMagnitude = 0;
  float shuffleUnbPx = 0;

  // Conflict graphs.
  // Nodes represent slits, and edge between two nodes indicates 
  // conflicting spectra between the two represented slits. 
  Graph allSlitsG;
  Graph acqSlitsG;
  Graph p1SlitsG;
  Graph p2SlitsG;
  Graph p3SlitsG;

  // Number of slits placed for each priority.
  int numberPlacedAcq;
  int numberPlacedPrio1;
  int numberPlacedPrio2;
  int numberPlacedPrio3;

  ofstream outStream;

  // Holds the microShuffle distance in pixels if in microshuffle mode.
  float microshufflePix = 0;

  // Load values from argv.
  if (argc >= 22) {
    strcpy(inFile, argv[1]);
    strcpy(outFileRoot, argv[2]);
    strcpy(fovFile, argv[3]);

    // spectrum length (pixels)
    specLength = atoi(argv[5]);

    // anamorphic factor
    anamorphic = atof(argv[6]);

    // central wavelength (nanometers)
    centralWavelength = (int) atof(argv[7]);

    // linear dispersion (nanometers / pixel)
    dPix = atof(argv[8]);

    // Lambdas 1 and 2. Used in spectrum calculations.
    l1Max = atof(argv[9]);
    l2Min = atof(argv[10]);

    // pixel scale (arcseconds / pixel)
    pixelScale = atof(argv[11]);

    // number of ODF masks to construct
    numMasks = (int) atof(argv[12]);

    // Max or Normal mode ('M' or 'N')
    slitMode = argv[13][0];

    // instrument type ("F2" || "GMOS" || "GMOS-S" || "GMOS-N")
    instType = argv[15];

    strcpy(det_img, argv[16]);
    strcpy(det_spec, argv[17]);

    // Coordinates of the preimage used for mask making
    strcpy(ra_img, argv[18]);
    strcpy(dec_img, argv[19]);

    minpixsep = atof(argv[20]);

    // This value determines how much a slit can be moved along its
    // length-wise axis when attempting to 'wiggle' more slits onto the mask
    wiggleVal = atof(argv[21]);

    // Print some introductory information. 
    // In normal GMMPS operation this will be displayed in the "Design Mask"
    // window defined in gmmps_spoc.tcl.
    printIntro(slitMode, instType);

  } 
  else {
    // Too few input parameters.
    printIntroError(argc);
    return (-1);
  }

  if (argc >= 23 && argv[22][0] != 0) {
    strcpy(bdfile, argv[22]);
  }

  // determine the minimum split separation based on the instrument type:
  // 4 pixels minimum
  MIN_SPEC_DIST = minpixsep * pixelScale;

  // Load in slit data from temporary data file generated in gmmps_spoc.tcl
  microshufflePix = loadSlits(&slits, inFile, fovFile, bdfile, pixelScale,
			      anamorphic, dPix, l1Max, l2Min, centralWavelength, instType,
			      wiggleVal, &bandLocations, &shuffleMagnitude, &shuffleUnbPx);

  if (microshufflePix == -1.0) {
    // Error case for nod and shuffle mode... tilted slit in nod and shuffle.
    printf(
	   "INVALID OBJECT TABLE.\nTILTED SLITS NOT ALLOWED IN N&S MASK DESIGN.\nODF NOT CREATED.\n");
    return 0;
  }

  // Search slit container for acquisition slits.
  acqSlits = getSlits(slits, '0');

  // Select slits for each mask up to numMasks.
  for (int i = 0; i < numMasks; i++) {
    numberPlacedAcq = 0;
    numberPlacedPrio1 = 0;
    numberPlacedPrio2 = 0;
    numberPlacedPrio3 = 0;

    // Init output stream.
    sprintf(outFile, "%s%d.cat", outFileRoot, i + 1);
    outStream.open(outFile);

    char outShortFile[128];
    if (strrchr(outFile, '/') != 0)
      sprintf(outShortFile, "%s", strrchr(outFile, '/') + 1);
    else
      sprintf(outShortFile, "%s", outFile);

    printf("\n\nMASK %d (of %d): %s\n", i + 1, numMasks, outShortFile);
    printf("  Objects available: %d\n", int(slits.size()));

    // Write ODF header.
    initOutputFile(outFile, outStream, bdfile, specLength, instType,
		   pixelScale, det_img, det_spec, ra_img, dec_img, shuffleUnbPx);

    // Acquisition slit processing.

    // Generate conflict map using only acquisition slits.
    mapConflicts(&acqSlitsG, acqSlits, specLength, microshufflePix,
		 instType, pixelScale);

    // Place as many acquisition slits as we can.
    placeSlits(&acqSlits, &placed, &removed, &acqSlits, &acqSlitsG,
	       specLength, microshufflePix, instType, pixelScale, fovFile);
    numberPlacedAcq = placed.size();

    // Generate conflict map using all slits.
    mapConflicts(&allSlitsG, slits, specLength, microshufflePix, instType,
		 pixelScale);

    // Remove any slits from "slits" which have been placed and add them to
    // "removed".
    removeConflicts(&slits, &placed, &removed, allSlitsG);

    // Priority 1 slit processing.

    // Fetch all priority 1 slits.
    p1Slits = getSlits(slits, '1');

    // Generate conflict graph using priority 1 slits only.
    mapConflicts(&p1SlitsG, p1Slits, specLength, microshufflePix, instType,
		 pixelScale);

    // Place as many priority 1 slits as we can.
    placeSlits(&slits, &placed, &removed, &p1Slits, &p1SlitsG, specLength,
	       microshufflePix, instType, pixelScale, fovFile);
    numberPlacedPrio1 = placed.size() - numberPlacedAcq;

    // Remove slits which now cannot be placed from consideration.
    removeConflicts(&slits, &placed, &removed, allSlitsG);

    // Priority 2 slit processing.
    p2Slits = getSlits(slits, '2');
    mapConflicts(&p2SlitsG, p2Slits, specLength, microshufflePix, instType,
		 pixelScale);
    placeSlits(&slits, &placed, &removed, &p2Slits, &p2SlitsG, specLength,
	       microshufflePix, instType, pixelScale, fovFile);
    numberPlacedPrio2 = placed.size() - numberPlacedAcq - numberPlacedPrio1;
    removeConflicts(&slits, &placed, &removed, allSlitsG);

    // Priority 3 slit processing.
    p3Slits = getSlits(slits, '3');
    mapConflicts(&p3SlitsG, p3Slits, specLength, microshufflePix, instType,
		 pixelScale);
    placeSlits(&slits, &placed, &removed, &p3Slits, &p3SlitsG, specLength,
	       microshufflePix, instType, pixelScale, fovFile);
    numberPlacedPrio3 = placed.size() - numberPlacedAcq - numberPlacedPrio1
      - numberPlacedPrio2;
    removeConflicts(&slits, &placed, &removed, allSlitsG);

    // Adjust slit positions slightly to try to place more slits on the mask
    if (wiggleVal > 0) {
      wiggleUnplaced(&slits, &placed, &removed, specLength,
		     microshufflePix, instType, pixelScale);
    }

    // Expand slits length-wise into each other, to maximize sky.
    if (slitMode == 'M' && microshufflePix == 0) {
      // Max mode is set and we are not in microshuffle mode.
      maxSlitMode(placed, pixelScale, instType, fovFile);

      if (bdfile[0] != 0) {
	// Reduce slits back into their bands when in band shuffle mode.

	maxModeEnforceBands(placed, shuffleMagnitude, bandLocations,
			    instType, pixelScale);
      }
    }

    // Write slits to ODF file.
    writeSlits(placed, outStream);

    // Print placement summary to the 'Design Masks' window.
    printf("  Objects included : %d, of which\n", int(placed.size()));
    printf("  Acquisition:       %d\n", numberPlacedAcq);
    printf("  Priority 1:        %d\n", numberPlacedPrio1);
    printf("  Priority 2:        %d\n", numberPlacedPrio2);
    printf("  Priority 3:        %d\n", numberPlacedPrio3);
    if (numberPlacedAcq < 2) {
      printf("  WARNING: Less than 2 acquisition objects!!\n");
    }

    // If we're constructing another ODF file, rearrange data to do so.
    if (i + 1 != numMasks) {
      // Removed is a list of all unplaced slits. This becomes our new input data. 
      // NOTE : 'slits' container should be empty at this point. 
      slits.swap(removed);

      // Clear all other containers. 
      // NOTE: Except 'acqSlits', those slits are re-used for each mask.
      removed.clear();
      placed.clear();
      p1Slits.clear();
      p2Slits.clear();
      p3Slits.clear();
    }

    outStream.close();
  }

  return 0;
}

/*
************************************************************************
*+
* FUNCTION: maxModeEnforceBands
*
* RETURNS: n/a
*
* DESCRIPTION: When in 'max sky' slit expansion mode and band shuffling mode
*               some slits are likely to have been expanded into the empty
*               space between bands.
*
*               This function sets those slits' boundaries back to within the
*               band.
*
* [NOTES:]:
*
*-
************************************************************************
*/
void maxModeEnforceBands(map<int, Slit> &placed, int bandShuffleMagnitude,
			 vector<float> &bandLocations, string instType, float pixelScale) {

  if (bandLocations.size() == 0) {
    return;
  }

  int bandIndex = 0;
  float bandStart = bandLocations[bandIndex];
  float bandEnd = bandStart + bandShuffleMagnitude;

  map<int, Slit>::iterator slitMapIter;
  int slitId;
  float slitStart, slitEnd;
  float newSlitStart, newSlitEnd;
  char slitPriority;
  bool slitChanged;

  // These are used to signal situations where a slit falls within a band and
  // is valid, but can be expanded back towards the "start" of the band.
  bool lastSlitEndChanged = false;
  bool newBand = true;

  for (slitMapIter = placed.begin(); slitMapIter != placed.end();
       slitMapIter++) {
    slitChanged = false;
    slitId = (*slitMapIter).second.id;
    slitStart = (*slitMapIter).second.slitStart;
    slitEnd = (*slitMapIter).second.slitEnd;
    slitPriority = (*slitMapIter).second.priority;

    // Test if we should move on to the next band.
    if (slitStart > bandEnd) {
      bandIndex++;
      if ((unsigned) bandIndex >= bandLocations.size()) {
	// Out of bands... there shouldn't be anything more to do, but
	// go through the rest of the slits to be sure.
	continue;
      }
      bandStart = bandLocations[bandIndex];
      bandEnd = bandStart + bandShuffleMagnitude;
      newBand = true;
    } 
    else {
      newBand = false;
    }

    // We never edit acquisition slits.
    if (slitPriority == '0') {
      lastSlitEndChanged = false;
      continue;
    }

    // Check if the slit has been expanded into a non-band region.
    if ((slitStart < bandStart && slitEnd > bandStart)
	|| (lastSlitEndChanged && newBand) // Or valid but can expand.
	) {
      newSlitStart = bandStart + MIN_SPEC_DIST;
      placed[slitId].slitStart = newSlitStart;
      slitChanged = true;
    }
    if (slitStart < bandEnd && slitEnd > bandEnd) {
      newSlitEnd = bandEnd - MIN_SPEC_DIST;
      placed[slitId].slitEnd = newSlitEnd;
      slitChanged = true;

      lastSlitEndChanged = true;
    } 
    else {
      lastSlitEndChanged = false;
    }

    // Recalculate slit length, spectrum position, etc.
    if (slitChanged) {
      maxOptProcessLine(placed, slitId, instType, pixelScale);
    }
  }
}

/*
************************************************************************
*+
* FUNCTION: initOutputFile
*
* RETURNS: n/a
*
* DESCRIPTION: Writes some preliminary data to the output file.  
*
* [NOTES:]: 
*
*-
************************************************************************
*/
void initOutputFile(char * outFile, ofstream & outStream, char * bdfile,
		    float specLen, string instType, float pixelScale, char * det_img,
		    char * det_spec, char * ra_img, char * dec_img, float shuffleUnbPx) {
  char outShortFile[128];
  char bdShortFile[128];
  char temp[1024];

  if (strrchr(outFile, '/') != 0) sprintf(outShortFile, "%s", strrchr(outFile, '/') + 1);
  else sprintf(outShortFile, "%s", outFile);

  if (strrchr(bdfile, '/') != 0) sprintf(bdShortFile, "%s", strrchr(bdfile, '/') + 1);
  else sprintf(bdShortFile, "%s", bdfile);

  // This writes to the .cat file.
  outStream.write("QueryResult\n\n", strlen("QueryResult\n\n"));
  outStream.write("# Config entry for original catalog server:\n",
		  strlen("# Config entry for original catalog server:\n"));
  outStream.write("serv_type: local\n", strlen("serv_type: local\n"));
  sprintf(temp, "long_name: %s\n", outFile);
  outStream.write(temp, strlen(temp));
  sprintf(temp, "short_name: %s\n", outShortFile);
  outStream.write(temp, strlen(temp));
  sprintf(temp, "url: %s\n", outFile);
  outStream.write(temp, strlen(temp));
  sprintf(temp,
	  "symbol: {x_ccd y_ccd priority} {circle white {} {} {} {$priority == \"3\"}} {15 {}} : {x_ccd y_ccd priority} {triangle blue {} {} {} {$priority == \"1\"}} {15 {}} : {x_ccd y_ccd priority} {diamond cyan {} {} {} {$priority == \"0\"}} {20 {}} : {x_ccd y_ccd priority} {cross yellow {} {} {} {$priority == \"X\"}} {15 {}} : {x_ccd y_ccd priority} {square green {} {} {} {$priority == \"2\"}} {15 {}}\n");
  outStream.write(temp, strlen(temp));
  sprintf(temp, "#SpectrumLength: %d\n", (int) specLen);
  outStream.write(temp, strlen(temp));
  outStream.write("# Fits keywords\n", strlen("# Fits keywords\n"));
  sprintf(temp, "#fits INSTRUME = %s\n", instType.c_str());
  outStream.write(temp, strlen(temp));
  sprintf(temp, "#fits PIXSCALE = %f\n", pixelScale);
  outStream.write(temp, strlen(temp));
  sprintf(temp, "#fits DET_IMG  = %s\n", det_img);
  outStream.write(temp, strlen(temp));
  sprintf(temp, "#fits DET_SPEC = %s\n", det_spec);
  outStream.write(temp, strlen(temp));
  sprintf(temp, "#fits RA_IMG   = %s\n", ra_img);
  outStream.write(temp, strlen(temp));
  sprintf(temp, "#fits DEC_IMG  = %s\n", dec_img);
  outStream.write(temp, strlen(temp));

  if (bdfile[0] != 0) {
    sprintf(temp, "#fits BDFILE   = %s\n", bdShortFile);
    outStream.write(temp, strlen(temp));
    sprintf(temp, "#fits SHUFSIZE  = %d / Shuffle distance in unbinned pixels\n", int(shuffleUnbPx));
    outStream.write(temp, strlen(temp));
  }

  outStream.write("# End fits keywords\n", strlen("# End fits keywords\n"));
  outStream.write("# End config entry\n\n", strlen("# End config entry\n\n"));
  sprintf(temp,
	  "ID	RA	DEC	x_ccd	y_ccd	specpos_x	specpos_y	slitpos_x	slitpos_y	slitsize_x	slitsize_y	slittilt	MAG	priority	slittype	wave_ccd\n");
  outStream.write(temp, strlen(temp));
  sprintf(temp,
	  "------	---------	---------	---------	---------	-------	-------	------	------	---------	---------	--------	---	--------	--------	--------\n");
  outStream.write(temp, strlen(temp));
}

/*
************************************************************************
*+
* FUNCTION: loadSlits
*
* RETURNS: Zero if not in microShuffle mode, the microShuffle distance
*          (in arcseconds) otherwise.
*
* DESCRIPTION: Loads all slit data from the input file.
*
* [NOTES:]:
*-
************************************************************************
*/
float loadSlits(map<int, Slit> *slits, char *inFileName, char *fovFile,
		char *bdFile, float pixelScale, float anamorphic, float dPix,
		float l1max, float l2min, float cWavelength, string instType,
		float wiggleFactor, vector<float> *shuffleBands,
		float* shuffleMagnitude, float* shuffleUnbPx) {

  ifstream inFile;

  // Holds fov vars {WDIM, WBEGIN, WSTOP, WCORNER1, WCORNER2, LDIM, LBEGIN, LSTOP, LCORNER1, LCORNER2}
  vector<float> fovSpecs(10);
  float WDIM;

  // Nod and Shuffle variables.
  char parmline[1024];    // Input buffer. 
  char key[1024];         // Parameter key buffer.
  char strval[256];       // Parameter value buffer (string).
  float val = 0.0;        // Parameter value buffer (float).
  int ival;               // Parameter value buffer (int).
  float binning = 0.0;    // Image binning value. 
  float bandLength = 0.0; // The size of each band in pixels.
  float shuffleAmt = 0.0;
  float shufflePix = 0.0; // The magnitude of the shuffling, in pixels. 
  float nodPix;           // The magnitude of nodding, in pixels.
  float msSlitLen = 0.0;
  bool bandShuffle = false;  // bandShuffle mode flag.
  bool microShuffle = false; // microShuffle mode flag. 

  // Slit data variables.

  // Useful vars for calculations:
  int id;
  float ccdW, ccdL;            // width-wise and length-wise ccd slit position
  float slitOffsetW, slitOffsetL;
  float slitLength, slitWidth;
  char priority;
  float slitStart; // The length-wise pixel coordinate of the start of the slit
  float slitEnd;   // The length-wise pixel coordinate of the end of the slit
  float slitTop, slitBottom;
  float wavePos;   // The ccd coordinate of the central wavelength. 
  string line;
  char lineChar[512];

  // These we only need because we need to output them later:
  float ra, dec, angle, mag;
  char type;
  float specPosW, specPosL; // these two will need to be calculated. 
  float wlambdac, lamda1, lamda2, w1, w2;

  // Load field of view information.
  fovSpecs = loadFov(fovFile, pixelScale, instType);
  WDIM = fovSpecs[0];

  // Nod and shuffle initiation.
  if (bdFile[0] != 0) {
    FILE *bdf;
    bdf = fopen(bdFile, "r");

    if (bdf != NULL) {

      // Input data from band definition file.
      while (fgets(parmline, sizeof(parmline), bdf) != NULL) {
	sscanf(parmline, "%s %s", key, strval);
	val = (float) atof(strval);
	ival = (int) atoi(strval);

	if (!strcmp("band:", key)) {
	  // NOTE: 'band:' info must be read in after binning value
	  //   is loaded.

	  int bandID;
	  char bandName[1024];
	  float bandPos;
	  float bandSz;

	  sscanf(parmline, "%s %d %s %f %f", key, &bandID, bandName,
		 &bandPos, &bandSz);

	  bandPos = bandPos / binning;

	  shuffleBands->push_back(bandPos);
	} 
	else if (!strcmp("shuffleMode:", key)) {
	  // Determine "bandShuffle" or 'microShuffle" mode. 

	  if (strcmp("bandShuffle", strval) == 0) {
	    bandShuffle = true;
	  } 
	  else if (strcmp("microShuffle", strval) == 0) {
	    microShuffle = true;
	  } 
	  else {
	    printf("Error! shuffle mode not properly set (%s)",
		   strval);
	    exit(0);
	  }
	} 
	else if (!strcmp("binning:", key)) {
	  // Set binning.
	  binning = val;
	} 
	else if (!strcmp("slitlen:", key)) {
	  // Determine the size of each band (or slit in microshuffle
	  // mode).

	  bandLength = val; // for use if in bandShuffle mode.
	  msSlitLen = val; // microShuffle mode. 
	} 
	else if (!strcmp("shufflePx:", key)) {
	  // NOTE: value is in unbinned pixels here
	  //  but below I convert it to pixelscale (binned) pixels
	  // don't round... value already rounded
	  // in band_def_UI
	  *shuffleUnbPx = val;
	} 
	else if (!strcmp("shuffleAmt:", key)) {
	  // value is in arcsec
	  shuffleAmt = val;
	} 
	else if (!strcmp("nodAmt:", key)) {
	  // convert to pixels from arcsec
	  nodPix = val / pixelScale;
	}
      } // End parameter load loop. 

      // Convert to _binned_ pixels
      shufflePix = shuffleAmt / pixelScale;
      // Set shuffle magnitude for access in 'main' function;
      *shuffleMagnitude = shufflePix;

      // shufflePix = shufflePix / binning;
      bandLength = bandLength / binning;

    }
  } // End nod/shuffle init. 

  inFile.open(inFileName, ios::in);
  if (inFile.is_open()) {
    // File has been opened correctly, begin data processing.

    vector < string > lineData;

    // Read slits from input file.
    while (!inFile.eof()) {
      getline(inFile, line);

      // Parse line data.
      lineData = stringSplit(line, " ");
      if (lineData.size() == 13) {
	stringToInt(lineData[0], id);
	stringToFloat(lineData[1], ra);
	stringToFloat(lineData[2], dec);
	stringToFloat(lineData[9], angle);
	stringToFloat(lineData[10], mag);
	priority = lineData[11][0];
	type = lineData[12][0];

	// Some fields are different based on instType.
	if (instType == "F2") {
	  /* Input order for F2:
	   * id (0),
	   * ra (1),
	   * dec(2),
	   * ccdL (3),
	   * ccdW (4),
	   * slitOffsetL (5),
	   * slitOffsetW (6),
	   * slitLength (7),
	   * slitWidth (8),
	   * angle (9),
	   * mag (10),
	   * priority (11),
	   * type (12)
	   */
	  stringToFloat(lineData[3], ccdL);
	  stringToFloat(lineData[4], ccdW);
	  stringToFloat(lineData[5], slitOffsetL);
	  stringToFloat(lineData[6], slitOffsetW);
	  stringToFloat(lineData[7], slitLength);
	  stringToFloat(lineData[8], slitWidth);

	} 
	else if (instType == "GMOS" || instType == "GMOS-S"
		   || instType == "GMOS-N") {
	  /* Input order for GMOS:
	   * id (0),
	   * ra (1),
	   * dec(2),
	   * ccdW (3),
	   * ccdL (4),
	   * slitOffsetW (5),
	   * slitOffsetL (6),
	   * slitWidth (7),
	   * slitLength (8),
	   * angle (9),
	   * mag (10),
	   * priority (11),
	   * type (12)
	   */
	  stringToFloat(lineData[3], ccdW);
	  stringToFloat(lineData[4], ccdL);
	  stringToFloat(lineData[5], slitOffsetW);
	  stringToFloat(lineData[6], slitOffsetL);
	  stringToFloat(lineData[7], slitWidth);
	  stringToFloat(lineData[8], slitLength);
	}

	// Slit width is only half what it should be.
	// NOTE: Why is not determined.
	if (instType == "F2") {
	  slitWidth *= 2;
	} 
	else {
	  slitLength *= 2;
	}

	// Slits have a constant length if we are in microshuffle mode.
	if (microShuffle && priority != '0') {
	  slitLength = msSlitLen;
	}

	// Acq Objects have to be 2.0 x 2.0 arcseconds.
	if (priority == '0') {
	  slitLength = 2.0;
	  slitWidth = 2.0;
	}

	// Spectrum position calculations.
	wlambdac = WDIM / 2
	  - (WDIM / 2 - (ccdW + slitOffsetW / pixelScale))
	  / anamorphic;
	lamda1 = cWavelength - (WDIM - wlambdac) * dPix;
	lamda1 = (l1max > lamda1) ? l1max : lamda1;
	lamda2 = cWavelength + wlambdac * dPix;
	lamda2 = (l2min < lamda2) ? l2min : lamda2;
	w1 = wlambdac + (cWavelength - lamda1) / dPix;
	w2 = wlambdac - (lamda2 - cWavelength) / dPix;
	specPosW = (w1 + w2) / 2 - ccdW;      // Center of spectrum in w

	wavePos = wlambdac;
	specPosL = slitOffsetL / pixelScale;

	if (instType == "F2") {
	  // Output order is also different for each instrument. F2 order:
	  // id, ra, dec, ccdL, ccdW, specPosL, specPosW, slitOffsetL, slitOffsetW, slitLength, slitWidth, angle, mag, priority, type.
	  // ( 'line' is the what will be written to the ODF if this slit is chosen )
	  sprintf(lineChar,
		  "%6d\t%10.5f\t%10.5f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%c\t%c\t%8.6f\n",
		  id, ra, dec, ccdL, ccdW, specPosL, specPosW,
		  slitOffsetL, slitOffsetW, slitLength, slitWidth,
		  angle, mag, priority, type, wavePos);
	} 
	else if (instType == "GMOS" || instType == "GMOS-S"
		   || instType == "GMOS-N") {
	  //
	  // Output order is also different for each instrument. F2 order:
	  // id, ra, dec, ccdW, ccdL, specPosW, specPosL, slitOffsetW, slitOffsetL, slitWidth, slitLength, angle, mag, priority, type.
	  // ( 'line' is the what will be written to the ODF if this slit is chosen )
	  sprintf(lineChar,
		  "%6d\t%10.5f\t%10.5f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%8.6f\t%c\t%c\t%8.6f\n",
		  id, ra, dec, ccdW, ccdL, specPosW, specPosL,
		  slitOffsetW, slitOffsetL, slitWidth, slitLength,
		  angle, mag, priority, type, wavePos);
	}

	// convert char* to string for storage.
	line = lineChar;

	// Convert all slit dimensions into pixels.
	slitOffsetL /= pixelScale;
	slitOffsetW /= pixelScale;
	slitLength /= pixelScale;
	slitWidth /= pixelScale;

	// Calculate some more slit dimensions.
	ccdL = ccdL + slitOffsetL;
	ccdW = ccdW + slitOffsetW;
	slitStart = ccdL - (slitLength / 2.0);
	slitEnd = ccdL + (slitLength / 2.0);
	slitTop = ccdW + (slitWidth / 2.0);
	slitBottom = ccdW - (slitWidth / 2.0);

	// Add the slit to the slit list. 
	// If in band shuffling mode make sure the
	// slit is in a band.
	if ((!bandShuffle || bandShuffleCheck(shuffleBands, bandLength, slitLength, ccdL))) {
	  slits->insert(pair<int, Slit>(id, 
					Slit(id, priority, slitStart,
					     slitEnd, slitLength, ccdL, 
					     ccdW, slitWidth, slitTop, 
					     slitBottom, specPosL, 
					     specPosW, angle, wiggleFactor,
					     line)));
	}
      }
    }

    inFile.close();

    printf("Total number of objects available: %d\n", int(slits->size()));
  } 
  else {
    printf("Unable to open input file. Exiting.\n");
    exit(1);
  }

  // Return microshuffle distance value if in microshuffle mode.
  if (microShuffle) {
    return shufflePix;
  } 
  else {
    return 0;
  }
}

/*
************************************************************************
*+
* FUNCTION: getSlits
*
* RETURNS: A vector<Slit> of Slit objects
*
* DESCRIPTION: Scans vector<Slit> 'slitData' for slits of 'priority',
*               and returns a vector<Slit> of those found objects.
*
* [NOTES:]: O(n)
*-
************************************************************************
*/
map<int, Slit> getSlits(map<int, Slit> slitData, char priority) {
  int idCur;
  char prioCur;
  map<int, Slit> foundSlits;
  map<int, Slit>::iterator it; // Iterator object.

  // Loop through 'slit' container to find slits of the given prio.
  for (it = slitData.begin(); it != slitData.end(); it++) {
    idCur = (*it).second.id;
    prioCur = (*it).second.priority;

    if (prioCur == priority) {
      // Found a slit, insert into 'foundSlits' container. 
      foundSlits.insert(pair<int, Slit>((*it).first, Slit((*it).second)));
    }
  }

  return foundSlits;
}

/*
************************************************************************
*+
* FUNCTION: mapConflicts
*
* RETURNS: A Graph object conflict map.
*
* DESCRIPTION: Maps all conflicts in 'slitData' into 'conflicts' Graph.
*
* [NOTES:]: Two slits conflict when the spectra of these two slits 
*            overlap. Conflicts between slits are represented in the 
*            graph edges between the conflicting slits.
*            O(n^2) <
*-
************************************************************************
*/
void mapConflicts(Graph * slitConflicts, map<int, Slit> slitData,
		  float specLength, float microshufflePix, string instType,
		  float pixelScale) {

  map<int, Slit>::iterator slitOne; // Iterator object.
  map<int, Slit>::iterator slitTwo;

  // Add nodes to conflict graph.
  for (slitOne = slitData.begin(); slitOne != slitData.end(); slitOne++) {
    slitConflicts->addNode((*slitOne).second.id);
  }

  // Search for conflicts between slits.
  for (slitOne = slitData.begin(); 1; slitOne++) {

    // Point slitTwo iterator at the position after slitOne.
    slitTwo = slitData.begin();
    while ((*slitTwo).first <= (*slitOne).first && slitTwo != slitData.end()) {
      slitTwo++;
    }

    // This means we're at the end of the data, exit loop.
    if (slitTwo == slitData.end())
      break;

    // See if data at slitOne conflicts with the remaining slits.
    for (; slitTwo != slitData.end(); slitTwo++) {
      if (conflicts((*slitOne).second, (*slitTwo).second, specLength,
		    microshufflePix, instType, pixelScale)) {
	slitConflicts->addEdge(slitConflicts->getNode((*slitOne).first),
			       slitConflicts->getNode((*slitTwo).first));
      }
    }
  }
}

/*
************************************************************************
*+
* FUNCTION: placeSlits
*
* RETURNS: none.
*
* DESCRIPTION: Select slits to place on the ODF. 
*
* [NOTES:]:
*-
************************************************************************
*/

void placeSlits(map<int, Slit> * slits, map<int, Slit> * placed,
		map<int, Slit> * removed, map<int, Slit> * current,
		Graph * conflictGraph, float specLength, float microshufflePix,
		string instType, float pixelScale, char * fovFile) {
  GraphNode * slitNode;

  vector<int> conflictList;
  vector<int> minDegSlits;
  vector<int>::iterator it;
  vector<float> fovSpecs(10);
  map<int, Slit>::iterator slitIterator;
  Slit slit;
  int sid;
  float fovStart;
  float fovEnd;
  float fovMid;

  // Find Field of view center for use later. 
  // fovSpects holds fov vars {WDIM, WBEGIN, WSTOP, WCORNER1, WCORNER2, LDIM, LBEGIN, LSTOP, LCORNER1, LCORNER2}
  fovSpecs = loadFov(fovFile, pixelScale, instType);
  fovStart = fovSpecs[6];
  fovEnd = fovSpecs[7];
  fovMid = ((fovEnd - fovStart) / 2 + fovStart) / pixelScale;

  // Place slits until conflict graph is empty.
  // NOTE: this could be more efficient. 
  while (!conflictGraph->empty()) {
    // Gather cardinality list.
    minDegSlits = conflictGraph->getMinDegree();

    // Find slit with max spectrum.
    sid = findMaxSpectrumSid(&minDegSlits, slits, fovMid);

    (*placed)[sid] = (*slits)[sid];

    // Remove conflicting slits.
    slitNode = conflictGraph->getNode(sid);
    conflictList = (*slitNode).getList();
    for (it = conflictList.begin(); it != conflictList.end(); it++) {
      conflictGraph->removeNode((*it));
      removeSlit((*it), slits, removed);
    }

    // Remove references to the placed node within the conflictGraph.
    conflictGraph->removeNode(sid);

    // Try to cozy the placed slit into its neighbors, if allowed.		
    if (wiggleNear(sid, slits, placed, pixelScale, instType)) {
      // If we've wiggled then update the conflict graph.

      mapConflicts(conflictGraph, (*current), specLength, microshufflePix,
		   instType, pixelScale);
    }
  }

  // After slits have been placed wiggle them a bit more.
  for (slitIterator = (*placed).begin(); slitIterator != (*placed).end();
       slitIterator++) {
    slit = (*slitIterator).second;
    wiggleNear(slit.id, slits, placed, pixelScale, instType);
  }
}

/*
************************************************************************
*+
* FUNCTION: checkConflicts
*
* RETURNS: True if there is a conflict, false otherwise.
*
* DESCRIPTION: Checks if a given slit conflicts with a given set of slits.
*
* [NOTES:]: 
*-
************************************************************************
*/
bool checkConflicts(Slit test, Slit except, map<int, Slit> slits,
		    float specLength, float msPix, string instType, float pixelScale) {
  map<int, Slit>::iterator slitIterator;
  Slit test2;

  for (slitIterator = slits.begin(); slitIterator != slits.end();
       slitIterator++) {
    test2 = (*slitIterator).second;
    if (test2.id != except.id
	&& conflicts(test, test2, specLength, msPix, instType,
		     pixelScale)) {
      return true;
    }
  }

  return false;
}

/*
************************************************************************
*+
* FUNCTION: maxSlitMode
*
* RETURNS: n/a
*
* DESCRIPTION: If in 'Max_Opt' mode, we expand slits to fill empty 
*              space, this is what this function does. 
*
* [NOTES:]:
*-
************************************************************************
*/
void maxSlitMode(map<int, Slit> &placed, float pixelScale, string instType,
		 char * fovFile) {
  vector<float> fovSpecs;
  map<float, int> sorted;
  map<int, Slit>::iterator slit;
  map<int, Slit>::iterator slit2;
  map<float, int>::iterator it;
  map<float, int>::iterator it2;
  map<float, int>::iterator it3;
  float end;
  float start;
  fovSpecs = loadFov(fovFile, pixelScale, instType);

  // Make sure no two slits have the same slitStart value.
  for (slit = placed.begin(); slit != placed.end(); slit++) {
    for (slit2 = slit; slit2++, slit2 != placed.end();) {
      if ((*slit).second.slitStart == (*slit2).second.slitStart) {
	// Increment slitStart marginally.
	(*slit).second.slitStart += 0.00001;
      }
    }
  }

  // Populate 'sorted':
  for (slit = placed.begin(); slit != placed.end(); slit++) {
    sorted.insert(
		  pair<float, int>((*slit).second.slitStart, (*slit).first));
  }

  // Expand slits.
  int id1, id2;
  for (it = sorted.begin(); it2 = it, it2++, it2 != sorted.end(); it++) {
    // Pointers to slits in the 'sorted' vector.
    id1 = (*it).second;
    id2 = (*it2).second;
    it3 = it2;
    it3++;

    if (it == sorted.begin()) {
      // Starting case.

      // Expand first slit to FOV.
      if (placed[id1].priority != '0' && placed[id1].angle == 0.0) {

	expandSlitToFOV(placed, fovSpecs, id1, 's');
	end = placed[id2].slitStart - MIN_SPEC_DIST / pixelScale;

	if (end > placed[id1].slitEnd) {
	  placed[id1].slitEnd = end;
	  maxOptProcessLine(placed, id1, instType, pixelScale);
	}

      } 
      else if (placed[id2].priority != '0'
		 && placed[id2].angle == 0.0) {
	// Slit 1 is an acq object, expand neighbor.

	start = placed[id1].slitEnd + MIN_SPEC_DIST / pixelScale;
	if (start < placed[id2].slitStart) {
	  placed[id2].slitStart = start;
	  maxOptProcessLine(placed, id2, instType, pixelScale);
	}
      }

    } 
    else if (it3 == sorted.end()) {
      // End case. 
      //
      // In this case 'it' points to the second last object, and 'it2'
      // points to the last object. 

      // Expand 'it2' as much as possible (to FOV and towards the slit 
      // pointed to by 'it')

      // Process previous slit. 
      maxOptProcessLine(placed, id1, instType, pixelScale);

      if (placed[id2].priority != '0' && placed[id2].angle == 0.0) {
	expandSlitToFOV(placed, fovSpecs, id2, 'e');

	// Expand last slit into previous slit (to avoid more fov-corner-centric calcs).
	start = placed[id1].slitEnd + MIN_SPEC_DIST / pixelScale;
	if (start < placed[id2].slitStart) {
	  placed[id2].slitStart = start;
	  maxOptProcessLine(placed, id2, instType, pixelScale);
	}

      } 
      else if (placed[id1].priority != '0'
		 && placed[id1].angle == 0.0) {
	// The last slit is an acquisition slit, do not expand it.

	end = placed[id2].slitStart - MIN_SPEC_DIST / pixelScale;
	if (end > placed[id1].slitEnd) {
	  placed[id1].slitEnd = end;
	  maxOptProcessLine(placed, id1, instType, pixelScale);
	}
      }

    } 
    else if (placed[id1].priority != '0' && placed[id1].angle == 0.0) {
      // Normal Case, expand slit into the following slit in 'sorted'.
      end = placed[id2].slitStart - MIN_SPEC_DIST / pixelScale;
      if (end > placed[id1].slitEnd) {
	placed[id1].slitEnd = end;

	// Process changes to the slit.
	maxOptProcessLine(placed, id1, instType, pixelScale);
      }
    } 
    else if (placed[id2].priority != '0' && placed[id2].angle == 0.0) {
      // If the current slit is an acquisition slit or a tilted slit
      // expand the next slit in the sorted vector into it.
      start = placed[id1].slitEnd + MIN_SPEC_DIST / pixelScale;

      if (start < placed[id2].slitStart) {
	placed[id2].slitStart = start;
	maxOptProcessLine(placed, id2, instType, pixelScale);
      }
    }
  }
}

/*
************************************************************************
*+
* FUNCTION: expandSlitToFOV
*
* RETURNS: none.
*
* DESCRIPTION: Expands slitlength of a given slit to the Field of view.
*
* [NOTES:]: Used in Max Mode.
*-
************************************************************************
*/

void expandSlitToFOV(map<int, Slit> &placed, vector<float> &fov, int id,
		     char mode) {
  // Unpack FOV.
  // fov vector order : 0-WDIM, 1-WSTART, 2-WEND, 3-WCORNER1, 4-WCORNER2, 5-LDIM, 6-LSTART, 7-LEND, 8-LCORNER1, 9-LCORNER2
  float WSTART = fov[1];
  float WEND = fov[2];
  float WCORNER1 = fov[3];
  float WCORNER2 = fov[4];
  float LSTART = fov[6];
  float LEND = fov[7];
  float LCORNER1 = fov[8];
  float LCORNER2 = fov[9];

  // The positions of the relevant corners of the slit.
  float posW = placed[id].ccdW;
  float wOffset = 0; //placed[id].slitWidth / 2.0; // CHANGE LATER.

  // Expand slits.
  if ((posW + wOffset < WEND - WCORNER2)
      && (posW - wOffset > WSTART + WCORNER1)) {
    // Case 1 : Slit is in 'flat' part of the fov.
    // Expand to LSTART or LEND, depending on mode.
    if (mode == 's') {
      placed[id].slitStart = LSTART;
    } 
    else if (mode == 'e') {
      placed[id].slitEnd = LEND;
    }
  } 
  else {
    // Case 2 : Corners.

    float m, w0, w1, l0, l1;
    float newPosL;

    if (mode == 's') {
      // 'Start' mode case.
      l0 = LSTART;
      l1 = LSTART + LCORNER2;
    } 
    else {
      // 'End' mode case.
      l0 = LEND;
      l1 = LEND - LCORNER1;
    }

    if (posW + wOffset > WEND - WCORNER2) {
      // 'Top' case.
      w0 = WEND - WCORNER2;
      w1 = WEND;
    } 
    else {
      // 'Bottom' case.
      w0 = WSTART + WCORNER1;
      w1 = WSTART;
    }

    m = (w1 - w0) / (l1 - l0);
    newPosL = l0 - (w0 - posW) / m;

    if (newPosL > LEND) {
      newPosL = LEND;
    } 
    else if (newPosL < 0) {
      newPosL = 0;
    }

    // Expand slit to FOV.
    if (mode == 's') {
      placed[id].slitStart = newPosL;
    } 
    else if (mode == 'e') {
      placed[id].slitEnd = newPosL;
    }
  }
}

/*
************************************************************************
*+
* FUNCTION: maxOptProcessLine
*
* RETURNS: n/a
*
* DESCRIPTION: Process a line that has been expanded. 'ccdL', as well as 
*              'slitLength' need to be recalculated. 
*
* [NOTES:]:
*-
************************************************************************
*/
void maxOptProcessLine(map<int, Slit> &placed, int id1, string instType,
		       float pixelScale) {
  // Unpack 'line'; recalc 'ccdL' and 'slitLength'; repack 'line'

  // Used for packing/unpacking each slit's line.
  string id, ra, dec, ccdW, specPosW, slitOffsetW;
  string slitWidth, angle, mag, prio, type, wave_ccd;
  float ccdL = 0.0, slitLength = 0.0, slitOffsetL = 0.0, specPosL = 0.0;
  vector < string > lineData;

  // Unpack line.
  lineData = stringSplit(placed[id1].line, "\t");

  id = lineData[0];
  ra = lineData[1];
  dec = lineData[2];
  angle = lineData[11];
  mag = lineData[12];
  prio = lineData[13];
  type = lineData[14];

  wave_ccd = "";

  if (lineData.size() > 15) {
    wave_ccd = lineData[15];

    // Remove trailing newline and whitespace.
    wave_ccd.erase(wave_ccd.find_last_not_of(" \t\r\n") + 1);
  } 
  else {
    type.erase(type.find_last_not_of(" \t\r\n") + 1);
  }

  // Some fields are different based on instType.
  if (instType == "F2") {
    stringToFloat(lineData[3], ccdL);
    ccdW = lineData[4];
    stringToFloat(lineData[5], specPosL);
    specPosW = lineData[6];
    stringToFloat(lineData[7], slitOffsetL);
    slitOffsetW = lineData[8];
    stringToFloat(lineData[9], slitLength);
    slitWidth = lineData[10];
  } 
  else if (instType == "GMOS" || instType == "GMOS-S" || instType == "GMOS-N") {
    ccdW = lineData[3];
    stringToFloat(lineData[4], ccdL);
    specPosW = lineData[5];
    stringToFloat(lineData[6], specPosL);
    slitOffsetW = lineData[7];
    stringToFloat(lineData[8], slitOffsetL);
    slitWidth = lineData[9];
    stringToFloat(lineData[10], slitLength);
  }

  // Done unpacking line. 
  // Now recalc specPosL (specpos_y), slitOffsetL (slitpos_y), slitLength. 
  slitLength = placed[id1].slitEnd - placed[id1].slitStart;
  specPosL = (placed[id1].slitEnd + placed[id1].slitStart) / 2. - ccdL;
  slitOffsetL = specPosL * pixelScale;
  slitLength *= pixelScale;

  // Repack line. 
  char lineChar[256];

  if (instType == "F2") {
    // Output order is also different for each instrument. F2 order:
    // id, ra, dec, ccdL, ccdW, specPosL, specPosW, slitOffsetL, slitOffsetW, slitLength, slitWidth, angle, mag, priority, type.
    // ( 'line' is the what will be written to the ODF if this slit is chosen )
    sprintf(lineChar,
	    "%s\t%s\t%s\t%8.6f\t%s\t%8.6f\t%s\t%8.6f\t%s\t%8.6f\t%s\t%s\t%s\t%s\t%s\t%s\n",
	    id.c_str(), ra.c_str(), dec.c_str(), ccdL, ccdW.c_str(),
	    specPosL, specPosW.c_str(), slitOffsetL, slitOffsetW.c_str(),
	    slitLength, slitWidth.c_str(), angle.c_str(), mag.c_str(),
	    prio.c_str(), type.c_str(), wave_ccd.c_str());
  } 
  else if (instType == "GMOS" || instType == "GMOS-S" || instType == "GMOS-N") {
    //
    // Output order is also different for each instrument. F2 order:
    // id, ra, dec, ccdW, ccdL, specPosW, specPosL, slitOffsetW, slitOffsetL, slitWidth, slitLength, angle, mag, priority, type.
    // ( 'line' is the what will be written to the ODF if this slit is chosen )
    sprintf(lineChar,
	    "%s\t%s\t%s\t%s\t%8.6f\t%s\t%8.6f\t%s\t%8.6f\t%s\t%8.6f\t%s\t%s\t%s\t%s\t%s\n",
	    id.c_str(), ra.c_str(), dec.c_str(), ccdW.c_str(), ccdL,
	    specPosW.c_str(), specPosL, slitOffsetW.c_str(), slitOffsetL,
	    slitWidth.c_str(), slitLength, angle.c_str(), mag.c_str(),
	    prio.c_str(), type.c_str(), wave_ccd.c_str());
  }

  placed[id1].line = lineChar;
}

/*
************************************************************************
*+
* FUNCTION: wiggleUnplaced
*
* RETURNS: Sucess code.
*
* DESCRIPTION: Tries to fit more slits on the mask by moving slits up or down.
*
* [NOTES:]:
*-
************************************************************************
*/
int wiggleUnplaced(map<int, Slit> * slits, map<int, Slit> * placed,
		   map<int, Slit> * removed, float specLength, float microshufflePix,
		   string instType, float pixelScale) {
  // Conflict graph between all slits.
  Graph conflictGraph;

  // The slits (placed and unplaced).
  map<int, Slit> allSlits;
  map<int, Slit>::iterator placedIterator;

  // The number of placed slits each removed slits conflicts with.
  map<int, int> conflictNum;

  // Used for holding conflict lists of loaded placed slits.
  vector<int> slitConflictList;

  Slit placedSlit;
  Slit cur;
  unsigned int i;
  int remId;

  // Possible wiggle room for the already placed slit.
  float placedWiggleRoom = 0;

  // Possible wiggle room for the unplaced slit.
  float curWiggleRoom = 0;

  // How much a placed slit has wiggled.
  float placedWiggleMag = 0;
  float curWiggleMag = 0;

  // True if each respective slits can be wiggled successfully.
  bool placedMove = false;
  bool curMove = false;

  allSlits.insert((*placed).begin(), (*placed).end());
  allSlits.insert((*removed).begin(), (*removed).end());

  // Map conflicts between all slits.
  mapConflicts(&conflictGraph, allSlits, specLength, microshufflePix,
	       instType, pixelScale);

  /*
   * Count how many times each removed slit conflicts with a placed slit.
   * We need to keep track of this because we are not going to place removed
   *   slits that conflict with more than one placed slit.
   */
  for (placedIterator = (*placed).begin(); placedIterator != (*placed).end();
       placedIterator++) {
    placedSlit = (*placedIterator).second;
    slitConflictList = (*conflictGraph.getNode(placedSlit.id)).getList();

    for (i = 0; i < slitConflictList.size(); i++) {
      remId = slitConflictList[i];

      if (conflictNum.count(remId) == 1) {
	conflictNum[remId]++;
      } 
      else {
	conflictNum[remId] = 1;
      }
    }
  }

  /*
   * Search for slits that can be moved, and move them.
   * NOTE: Only two slits are considered at a time.
   */
  for (placedIterator = (*placed).begin(); placedIterator != (*placed).end();
       placedIterator++) {

    placedSlit = (*placedIterator).second;

    // Find valid wiggle amount for this slit.
    if (placedSlit.wiggleUsed == false) {
      placedWiggleRoom = placedSlit.wiggleRoom;
    } 
    else {
      // We can only wiggle once, so no dice.
      placedWiggleRoom = 0;
    }

    // Find list of slits excluded by this slit.
    slitConflictList = (*conflictGraph.getNode(placedSlit.id)).getList();

    // Try to place this slit if possible.
    // NOTE: Should be taking priority into account here?
    for (i = 0; i < slitConflictList.size(); i++) {
      remId = slitConflictList[i];

      if (conflictNum[remId] > 1) {
	// Can't add slits with more than one conflict.
	continue;
      }

      cur = (*removed)[remId];
      curWiggleRoom = cur.wiggleRoom;

      // Determine how the removed slit conflicts with the placed slit.
      if (placedSlit.slitEnd < cur.slitEnd
	  && placedSlit.slitEnd > cur.slitStart
	  && cur.slitStart > placedSlit.slitStart) {
	// Removed slit conflicts with the 'end' edge of the placed slit.

	// Test if it can be wiggled.
	if (placedSlit.slitEnd - cur.slitStart
	    + MIN_SPEC_DIST / pixelScale < curWiggleRoom) {
	  // Can fit while wiggling only removed slit.
	  curWiggleMag = placedSlit.slitEnd - cur.slitStart
	    + MIN_SPEC_DIST / pixelScale;
	  cur.changePosition(0, curWiggleMag, pixelScale, instType);

	  // Make sure mask is still valid.
	  if (checkConflicts(cur, placedSlit, (*placed), specLength,
			     microshufflePix, instType, pixelScale)) {
	    // Movement causes conflict, undo wiggle.
	    curWiggleMag *= -1;
	    cur.changePosition(0, curWiggleMag, pixelScale,
			       instType);

	  } 
	  else {
	    // Movement does not cause conflict, flag for placement.
	    curMove = true;
	  }

	} 
	else if (placedSlit.slitEnd - cur.slitStart
		   + MIN_SPEC_DIST / pixelScale
		   < curWiggleRoom + placedWiggleRoom) {
	  // Must wiggle both slits to fit.

	  // Determine relative wiggle magnitudes over both slits.
	  if (curWiggleRoom / 2.0
	      >= (placedSlit.slitEnd - cur.slitStart
		  + MIN_SPEC_DIST / pixelScale) / 2.0) {

	    if (placedWiggleRoom
		>= (placedSlit.slitEnd - cur.slitStart
		    + MIN_SPEC_DIST / pixelScale) / 2.0) {
	      // Can wiggle in equal parts.
	      curWiggleMag = (placedSlit.slitEnd - cur.slitStart
			      + MIN_SPEC_DIST / pixelScale) / 2.0;
	      placedWiggleMag = (placedSlit.slitEnd
				 - cur.slitStart + MIN_SPEC_DIST) / 2.0 * -1;

	    } 
	    else {
	      // Need more cur wiggle.
	      placedWiggleMag = placedWiggleRoom * -1;
	      curWiggleMag = placedSlit.slitEnd - cur.slitStart
		+ MIN_SPEC_DIST / pixelScale
		- placedWiggleRoom;
	    }

	  } 
	  else {
	    // Need more placed wiggle.
	    curWiggleMag = curWiggleRoom;
	    placedWiggleMag = (placedSlit.slitEnd - cur.slitStart
			       + MIN_SPEC_DIST / pixelScale - curWiggleRoom)
	      * -1;
	  }

	  placedSlit.changePosition(0, placedWiggleMag, pixelScale,
				    instType);
	  cur.changePosition(0, curWiggleMag, pixelScale, instType);

	  // Make sure mask is still valid.
	  if (checkConflicts(cur, placedSlit, (*placed), specLength,
			     microshufflePix, instType, pixelScale)
	      || checkConflicts(placedSlit, placedSlit, (*placed),
				specLength, microshufflePix, instType,
				pixelScale)) {
	    curWiggleMag *= -1;
	    cur.changePosition(0, curWiggleMag, pixelScale,
			       instType);
	    placedWiggleMag *= -1;
	    placedSlit.changePosition(0, placedWiggleMag,
				      pixelScale, instType);
	  } 
	  else {
	    curMove = true;
	    placedMove = true;

	  }
	}
      } 
      else if (placedSlit.slitStart > cur.slitStart
		 && placedSlit.slitStart < cur.slitEnd
		 && cur.slitEnd < placedSlit.slitEnd) {
	// Removed slit conflicts with the 'start' edge of the placed slit.

	// Test if slit can be wiggled.
	if (cur.slitEnd - placedSlit.slitStart
	    + MIN_SPEC_DIST / pixelScale < curWiggleRoom) {
	  // Can fit while wiggling only removed slit.

	  curWiggleMag = (cur.slitEnd - placedSlit.slitStart
			  + MIN_SPEC_DIST / pixelScale) * -1;
	  cur.changePosition(0, curWiggleMag, pixelScale, instType);

	  // Make sure mask is still valid.
	  if (checkConflicts(cur, placedSlit, (*placed), specLength,
			     microshufflePix, instType, pixelScale)) {
	    // Movement causes conflict, undo wiggle.

	    curWiggleMag *= -1;
	    cur.changePosition(0, curWiggleMag, pixelScale,
			       instType);
	  } 
	  else {
	    // Movement does not cause conflict, flag for placement.
	    curMove = true;
	  }

	} 
	else if (cur.slitEnd - placedSlit.slitStart
		   + MIN_SPEC_DIST / pixelScale
		   < curWiggleRoom + placedWiggleRoom) {
	  // Need to wiggle both slits to fit.

	  /*
	   * Determine relative wiggle magnitudes for the placed and
	   * unplaced slits.
	   */
	  if (curWiggleRoom / 2.0
	      >= (cur.slitEnd - placedSlit.slitStart
		  + MIN_SPEC_DIST / pixelScale) / 2.0) {

	    if (placedWiggleRoom
		>= (cur.slitEnd - placedSlit.slitStart
		    + MIN_SPEC_DIST / pixelScale) / 2.0) {
	      // Can wiggle in equal parts.
	      curWiggleMag = (cur.slitEnd - placedSlit.slitStart
			      + MIN_SPEC_DIST / pixelScale) / 2.0 * -1;
	      placedWiggleMag = (cur.slitEnd
				 - placedSlit.slitStart
				 + MIN_SPEC_DIST / pixelScale) / 2.0;
	    } 
	    else {
	      // Need more cur wiggle.
	      placedWiggleMag = placedWiggleRoom;
	      curWiggleMag = (cur.slitEnd - placedSlit.slitStart
			      + MIN_SPEC_DIST / pixelScale
			      - placedWiggleRoom) * -1;
	    }
	  } 
	  else {
	    // Need more placed wiggle.
	    curWiggleMag = curWiggleRoom * -1;
	    placedWiggleMag = cur.slitEnd - placedSlit.slitStart
	      + MIN_SPEC_DIST / pixelScale - curWiggleRoom;
	  }

	  placedSlit.changePosition(0, placedWiggleMag, pixelScale,
				    instType);
	  cur.changePosition(0, curWiggleMag, pixelScale, instType);

	  // Make sure mask is still valid.
	  if (checkConflicts(cur, placedSlit, (*placed), specLength,
			     microshufflePix, instType, pixelScale)
	      || checkConflicts(placedSlit, placedSlit, (*placed),
				specLength, microshufflePix, instType,
				pixelScale)) {
	    curWiggleMag *= -1;
	    cur.changePosition(0, curWiggleMag, pixelScale,
			       instType);
	    placedWiggleMag *= -1;
	    placedSlit.changePosition(0, placedWiggleMag,
				      pixelScale, instType);
	  } 
	  else {
	    curMove = true;
	    placedMove = true;
	  }
	}
      } 
      else {
	// Other cases, ignore.
	continue;
      }

      // Save changes into slit maps.
      if (curMove) {
	cur.wiggleUsed = true;
	(*removed).erase(cur.id);
	(*placed)[cur.id] = cur;
	(*slits)[cur.id] = cur;
      }
      if (placedMove) {
	placedSlit.wiggleUsed = true;
	(*placed)[placedSlit.id] = placedSlit;
	(*slits)[placedSlit.id] = placedSlit;
      }

      curMove = false;
      placedMove = false;
    }
  }

  return 0;
}

/*
************************************************************************
*+
* FUNCTION: wiggleNear
*
* RETURNS: None.
*
* DESCRIPTION: Create more space by moving placed slits closer to each other
*               (within the permitted wiggle space)
*
* [NOTES:]:
*-
************************************************************************
*/
bool wiggleNear(int sid, map<int, Slit> * slits, map<int, Slit> * placed,
		float pixelScale, string instType) {
  map<int, Slit>::iterator slitIterator;

  // Gather information on slit to wiggle.
  Slit slit = (*slits)[sid];

  Slit cur;
  float curWiggleRoom;

  float wiggleMag = 0;
  float wiggleMagSlit = 0;
  float wiggleMagCur = 0;

  // If wiggle has been used then don't wiggle.
  if (slit.wiggleUsed) {
    return 0;
  }

  // Try to wiggle.
  for (slitIterator = (*placed).begin(); slitIterator != (*placed).end();
       slitIterator++) {
    cur = (*slitIterator).second;

    if (cur.id == sid) {
      // Don't compare slit to itself.
      continue;
    }

    // If this cur hasn't been wiggled already then we can wiggle it here.
    if (cur.wiggleUsed == true) {
      curWiggleRoom = 0;
    } 
    else {
      curWiggleRoom = cur.wiggleRoom;
    }

    // Detect and execute wiggle, if able.
    if (slit.slitEnd < cur.slitStart
	&& slit.slitEnd + slit.wiggleRoom
	> cur.slitStart - curWiggleRoom) {
      // Check top edge of slit to bottom edge of cur.

      wiggleMag = cur.slitStart - slit.slitEnd
	- MIN_SPEC_DIST / pixelScale;

      // Determine slit movement magnitudes.
      if (curWiggleRoom >= wiggleMag / 2.0
	  && slit.wiggleRoom >= wiggleMag / 2.0) {
	// Both slits can meet in the middle.
	wiggleMagCur = wiggleMag / 2.0;
	wiggleMagSlit = wiggleMag / 2.0;
      } 
      else if (curWiggleRoom < wiggleMag / 2.0) {
	// cur cannot wiggle enough.
	wiggleMagCur = curWiggleRoom;
	wiggleMagSlit = wiggleMag - curWiggleRoom;
      } 
      else if (slit.wiggleRoom < wiggleMag / 2.0) {
	// slit cannot wiggle enough.
	wiggleMagSlit = slit.wiggleRoom;
	wiggleMagCur = wiggleMag - wiggleMagCur;
      }

      wiggleMagCur *= -1;

      // Apply changes to slit objects.
      slit.changePosition(0, wiggleMagSlit, pixelScale, instType);
      cur.changePosition(0, wiggleMagCur, pixelScale, instType);
      slit.wiggleUsed = true;
      cur.wiggleUsed = true;

      (*slits)[slit.id] = slit;
      (*placed)[slit.id] = slit;
      (*slits)[cur.id] = cur;
      (*placed)[cur.id] = cur;
    } 
    else if (slit.slitStart > cur.slitEnd
	       && slit.slitStart - slit.wiggleRoom
	       < cur.slitEnd + curWiggleRoom) {

      // Check bottom edge of slit to top edge of cur.
      wiggleMag = slit.slitStart - cur.slitEnd
	- MIN_SPEC_DIST / pixelScale;

      // Determine slit movement magnitudes.
      if (curWiggleRoom >= wiggleMag / 2.0
	  && slit.wiggleRoom >= wiggleMag / 2.0) {
	// Both slits can meet in the middle.
	wiggleMagCur = wiggleMag / 2.0;
	wiggleMagSlit = wiggleMag / 2.0;
      } 
      else if (curWiggleRoom < wiggleMag / 2.0) {
	// cur cannot wiggle enough.
	wiggleMagCur = curWiggleRoom;
	wiggleMagSlit = wiggleMag - curWiggleRoom;
      } 
      else if (slit.wiggleRoom < wiggleMag / 2.0) {
	// slit cannot wiggle enough.
	wiggleMagSlit = slit.wiggleRoom;
	wiggleMagCur = wiggleMag - wiggleMagCur;
      }

      wiggleMagSlit *= -1;

      // Apply changes to slit objects.
      slit.changePosition(0, wiggleMagSlit, pixelScale, instType);
      cur.changePosition(0, wiggleMagCur, pixelScale, instType);
      slit.wiggleUsed = true;
      cur.wiggleUsed = true;

      (*slits)[slit.id] = slit;
      (*placed)[slit.id] = slit;
      (*slits)[cur.id] = cur;
      (*placed)[cur.id] = cur;
    }
  }

  return 0;
}

/*
************************************************************************
*+
* FUNCTION: conflicts
*
* RETURNS: true or false
*
* DESCRIPTION: Returns true if the two slits conflict, false otherwise.
*
* [NOTES:]: msPix represents the movement in Microshuffle mode. 
*-
************************************************************************
*/
bool conflicts(Slit slitOne, Slit slitTwo, float specLength, float msPix,
	       string instType, float pixelScale) {
  float laserPad;

  // Set minimum slit separation based on instrument type. (in pixels)
  if (instType == "F2") {
    laserPad = 2.685;
  } 
  else {
    laserPad = 2.7;
  }

  // Test spectra overlap in slitLength-wise direction.
  if (((slitTwo.slitEnd > slitOne.slitStart - MIN_SPEC_DIST / pixelScale
	&& slitTwo.slitEnd < slitOne.slitEnd + MIN_SPEC_DIST / pixelScale)
       || (slitTwo.slitStart
	   > slitOne.slitStart - MIN_SPEC_DIST / pixelScale
	   && slitTwo.slitStart
	   < slitOne.slitEnd + MIN_SPEC_DIST / pixelScale)
       || (slitTwo.slitStart
	   < slitOne.slitStart + MIN_SPEC_DIST / pixelScale
	   && slitTwo.slitEnd
	   > slitOne.slitEnd - MIN_SPEC_DIST / pixelScale))
      || (msPix != 0.0
	  && ((slitTwo.slitEnd
	       > slitOne.slitStart - msPix
	       - MIN_SPEC_DIST / pixelScale
	       && slitTwo.slitEnd
	       < slitOne.slitEnd + msPix
	       + MIN_SPEC_DIST / pixelScale)
	      || (slitTwo.slitStart
		  > slitOne.slitStart - msPix
		  - MIN_SPEC_DIST / pixelScale
		  && slitTwo.slitStart
		  < slitOne.slitEnd + msPix
		  + MIN_SPEC_DIST / pixelScale)
	      || (slitTwo.slitStart
		  < slitOne.slitStart + msPix
		  + MIN_SPEC_DIST / pixelScale
		  && slitTwo.slitEnd
		  > slitOne.slitEnd - msPix
		  - MIN_SPEC_DIST / pixelScale)))) {

    // The slits overlap length-wise, now test if the spectra overlap width-wise.
    float s1SpectraEdge2 = slitOne.ccdW + slitOne.specPosW
      + specLength / 2.0;
    float s1SpectraEdge1 = slitOne.ccdW + slitOne.specPosW
      - specLength / 2.0;
    float s2SpectraEdge2 = slitTwo.ccdW + slitTwo.specPosW
      + specLength / 2.0;
    float s2SpectraEdge1 = slitTwo.ccdW + slitTwo.specPosW
      - specLength / 2.0;

    if ((s2SpectraEdge1 <= s1SpectraEdge2 + MIN_SPEC_DIST / pixelScale &&
	 s2SpectraEdge1 >= s1SpectraEdge1 - MIN_SPEC_DIST / pixelScale) || 
	(s2SpectraEdge2 >= s1SpectraEdge1 - MIN_SPEC_DIST / pixelScale && 
	 s2SpectraEdge2 <= s1SpectraEdge2 + MIN_SPEC_DIST / pixelScale) || 
	(s2SpectraEdge1 <= s1SpectraEdge1 + MIN_SPEC_DIST / pixelScale && 
	 s2SpectraEdge2 >= s1SpectraEdge2 - MIN_SPEC_DIST / pixelScale)) {
      // These slits conflict with eachother! 

      return true;
    }

  } 
  else if (((slitTwo.slitEnd < slitOne.slitStart && 
	     slitOne.slitStart - laserPad < slitTwo.slitEnd) || 
	    (slitTwo.slitStart > slitOne.slitEnd && 
	     slitOne.slitEnd   + laserPad > slitTwo.slitStart)) && 
	   ((slitTwo.slitTop < slitOne.slitTop + laserPad || 
	     slitTwo.slitBottom < slitOne.slitTop + laserPad) &&
	    (slitTwo.slitTop > slitOne.slitBottom - laserPad || 
	     slitTwo.slitBottom > slitOne.slitBottom - laserPad))) {
    return true;
  }

  return false;
}

/*
************************************************************************
*+
* FUNCTION: removeConflicts
*
* RETURNS: A vector<Slit> containing the Slits that were not removed.
*
* DESCRIPTION: Scan conflictGraph for Slits that conflict with slits in 
*               placedSlits, remove those slits from slitData.
*
* [NOTES:]: O(n*m)
*-
************************************************************************
*/
void removeConflicts(map<int, Slit> * slits, map<int, Slit> * placed,
		     map<int, Slit> * removed, Graph conflictGraph) {
  map<int, Slit>::iterator it;
  vector<int>::iterator conIt;
  vector<int> conflictList;

  // Loop through placed slits, removing conflicts.
  for (it = placed->begin(); it != placed->end(); it++) {
    // Retreive list of slits (thier ID's really) conflicting with placed slit.
    conflictList = conflictGraph.getNode((*it).first)->getList();

    // Loop through slits and remove. 
    for (conIt = conflictList.begin(); conIt != conflictList.end();
	 conIt++) {
      conflictGraph.removeNode((*conIt));
      removeSlit((*conIt), slits, removed);
    }

  }
}

/*
************************************************************************
*+
* FUNCTION: loadFov
*
* RETURNS: A vector containing field of view dimensions. 
*
* DESCRIPTION: Loads the data contained in the fov file.
*
* [NOTES:]:
*-
************************************************************************
*/
vector<float> loadFov(char * fovFile, float pixelScale, string instType) {
  vector<float> fov(10);
  char oneLine[256], x[124]; /* One line of input.              */
  float XDIM, XBEGIN, XSTOP, YDIM, YBEGIN, YSTOP; /* fov dimensions using x/y coords */
  float XCORNER1, XCORNER2, YCORNER1, YCORNER2; /* more fov dim's                  */
  float tmp; /* Tmp value for fov file.         */
  int num; /* Num of var. read from file.     */
  FILE *fv; /* Fov file ptr.                   */

  XDIM = XBEGIN = XSTOP = YDIM = YBEGIN = YSTOP = 0;
  XCORNER1 = XCORNER2 = YCORNER1 = YCORNER2 = 0;

  if (fovFile[0] == '\0' || (fv = fopen(fovFile, "r")) == NULL) {
    printf("Failed to open field of view file<%s>, exiting.\n", fovFile);
    exit(-1);
  }

  // Read in fov file.
  while (fgets(oneLine, 256, fv) != NULL) {
    if (oneLine[0] == '#' || strlen(oneLine) < 4)
      continue;
    if (oneLine[strlen(oneLine) - 1] != '\n')
      continue;
    
    if (strncmp(oneLine, "XDIM", 4) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	XDIM = tmp / pixelScale;
    } 
    else if (strncmp(oneLine, "YDIM", 4) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	YDIM = tmp / pixelScale;
    } 
    else if (strncmp(oneLine, "XEND", 4) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	XSTOP = tmp / pixelScale;
    } 
    else if (strncmp(oneLine, "XSTART", 6) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	XBEGIN = tmp / pixelScale;
    } 
    else if (strncmp(oneLine, "YEND", 4) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	YSTOP = tmp / pixelScale;
    } 
    else if (strncmp(oneLine, "YSTART", 6) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	YBEGIN = tmp / pixelScale;
    }
    else if (strncmp(oneLine, "XCORNER1", 8) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	XCORNER1 = tmp / pixelScale;
    }
    else if (strncmp(oneLine, "XCORNER2", 8) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	XCORNER2 = tmp / pixelScale;
    }
    else if (strncmp(oneLine, "YCORNER1", 8) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	YCORNER1 = tmp / pixelScale;
    }
    else if (strncmp(oneLine, "YCORNER2", 8) == 0) {
      if ((num = sscanf(oneLine, "%s %f", x, &tmp)) == 2)
	YCORNER2 = tmp / pixelScale;
    }
  }

  // Fill fov vector based on instrument type.
  if (instType == "GMOS" || instType == "GMOS-S" || instType == "GMOS-N") {
    fov[0] = XDIM;
    fov[1] = XBEGIN;
    fov[2] = XSTOP;
    fov[3] = XCORNER1;
    fov[4] = XCORNER2;

    fov[5] = YDIM;
    fov[6] = YBEGIN;
    fov[7] = YSTOP;
    fov[8] = YCORNER1;
    fov[9] = YCORNER2;
  } 
  else if (instType == "F2") {
    fov[0] = YDIM;
    fov[1] = YBEGIN;
    fov[2] = YSTOP;
    fov[3] = YCORNER1;
    fov[4] = YCORNER2;

    fov[5] = XDIM;
    fov[6] = XBEGIN;
    fov[7] = XSTOP;
    fov[8] = XCORNER1;
    fov[9] = XCORNER2;
  }

  /* fov vector order:
   *  WDIM,
   *  WSTART,
   *  WEND,
   *  WCORNER1,
   *  WCORNER2,
   *  LDIM,
   *  LSTART,
   *  LEND,
   *  LCORNER1,
   *  LCORNER2
   */
  return fov;
}


/*
************************************************************************
*+
* FUNCTION: removeSlit
*
* RETURNS: n/a
*
* DESCRIPTION: Removes a slit from the 'slit' container and copies it 
*              into the 'removed' container for possible later use. 
*
* [NOTES:]: 
*
*-
************************************************************************
*/
void removeSlit(int sid, map<int, Slit> * slits, map<int, Slit> * removed) {
  //
  // copy slit to 'removed' container, for possible future use. 
  removed->insert(pair<int, Slit>(sid, Slit((*slits)[sid])));

  //
  // delete slit from 'slits' container.
  slits->erase(sid);
}

/*
************************************************************************
*+
* FUNCTION: writeSlits
*
* RETURNS: n/a
*
* DESCRIPTION: Writes slit data in the 'placed' container to the output
*              file.  
*
* [NOTES:]:
*-
************************************************************************
*/
void writeSlits(map<int, Slit> &placed, ofstream &outStream) {

  map<int, Slit>::iterator itSlits;
  string line;

  //
  // Write placed slits to file...
  for (itSlits = placed.begin(); itSlits != placed.end(); itSlits++) {
    line = (*itSlits).second.line;

    outStream.write(line.c_str(), strlen(line.c_str()));
  }
}

/*
************************************************************************
*+
* FUNCTION: bandShuffleCheck
*
* RETURNS: bool -- True if the slit is within a valid band, false otherwise.
*
* DESCRIPTION: See retrun value.
*
* [NOTES:]:
*-
************************************************************************
*/
bool bandShuffleCheck(vector<float> *shuffleBands, float bandLength,
		      float slitLength, float ccdL) {
  vector<float>::iterator bands;
  float slitStart, slitEnd, bandStart, bandEnd;

  slitStart = ccdL - slitLength / 2;
  slitEnd = ccdL + slitLength / 2;

  for (bands = shuffleBands->begin(); bands != shuffleBands->end(); bands++) {

    bandStart = (*bands);
    bandEnd = bandStart + bandLength;

    if (slitStart > bandStart && slitEnd < bandEnd) {
      return true;
    }
  }

  return false;
}

/*
************************************************************************
*+
* FUNCTION: findMaxSpectrumSid
*
* RETURNS: The slit id of the slit with the most spectral coverage on the CCD.
*
* DESCRIPTION: Given a list of slit ids and a vector of slit objects return
*  the one which is most likely to have the most of its spectrum fall on the
*  detector.
*
* [NOTES:]:
*-
************************************************************************
*/
int findMaxSpectrumSid(vector<int> * slitIds, map<int, Slit> * slits,
		       float fovMid) {
  int sid = 0;
  unsigned int i;
  float minDist = 32000;
  Slit cur;
  float curSpec;
  float curDist;
  int curId;

  // Find the slit whose spectrum is closest to being in the center of the
  // field of view.
  for (i = 0; i < (*slitIds).size(); i++) {
    curId = (*slitIds)[i];
    cur = (*slits)[curId];
    curSpec = cur.ccdL + cur.specPosL;
    curDist = (int) abs(fovMid - curSpec);

    if (curDist < minDist) {
      minDist = curDist;
      sid = curId;
    }
  }

  return sid;
}

/*
************************************************************************
*+
* FUNCTION: printIntro
*
* RETURNS: void
*
* DESCRIPTION: Print some introductory mask information to stdout.
*
* [NOTES:]: 
*
*-
************************************************************************
*/
void printIntro(char slitMode, string instType) {

  printf("----------------------------------------------------\n");
  printf("SPOC: SLIT POSITIONING OPTIMIZATION CODE\n");
  printf("INSTRUMENT MODE    : %s\n", instType.c_str());

  if (slitMode == 'M') {
    printf("SLIT SELECTION MODE: Max Sky\n\n");
  } 
  else {
    printf("SLIT SELECTION MODE: Normal\n\n");
  }
}

void printIntroError(int argc) {
  //Invalid input params.
  printf("----------------------------------------------------\n");
  printf("SPOC: SLIT POSITIONING OPTIMIZATION CODE\n\n");
  printf("ERROR: Invalid command line arguments: EXIT\n\n");
  printf(
	 "Usage: gmMakeMasks <inFilename> <outFilename> <slitWidth> <specLen> <anamorphic>, argc=%d\n",
	 argc);
  printf(
	 "\t<centralWavelength> <linear dispersion> <lamda1Max> <lamda2Min> \n");
  printf("\t<pixelScale> <numMasks> <option> <debugLevel> \n");
  printf("DETAILS:\nPARM 1:input file name\n");
  printf("PARM 2:output file name without extention\n");
  printf("PARM 3:field of view config filename\n");
  printf("PARM 4:slitWidth(pix), 0 indicates used input file values\n");
  printf("PARM 5:spectra length(pix)\n");
  printf("PARM 6:spectra anamorphic factor\n");
  printf("PARM 7:central wavelength\n");
  printf("PARM 8:Linear dispersion\n");
  printf("PARM 9:Lamda 1 max\n");
  printf("PARM 10:Lamda 2 min\n");
  printf("PARM 11:pixelScale conversion factor\n");
  printf("PARM 12:number of masks\n");
  printf("PARM 13:type of algorithm:N/n=Normal_Opt M/m=Max_Opt\n");
  printf("    N=do not change object X-dimension distribution\n");
  printf("    M=get more objects but small ones are favored\n");
  printf("PARM 14:Debug level, 0=none, to 3=max\n");
  printf("PARM 15:Instrument Type, i.e. GMOS or F2 (FLAMINGOS-2)\n");
  printf("PARM 16:DET_IMG (detector ID)\n");
  printf("PARM 17:DET_SPEC (detector ID)\n");
  printf("PARM 18:RA of the preimage\n");
  printf("PARM 19:DEC of the preimage\n");
  printf("----------------------------------------------------\n");
}

/*
************************************************************************
*+
* FUNCTION: stringSplit
*
* RETURNS: vector<string>
*
* DESCRIPTION: Splits the string by one or more instances of 'delim'.
*
* [NOTES:]: I found this algorithm on an internet message board. 
*
*-
************************************************************************
*/
vector<string> stringSplit(string str, string delim) {
  vector < string > results;

  int cutAt;
  while (int((cutAt = str.find_first_of(delim))) != int(str.npos)) {
    if (cutAt > 0) {
      results.push_back(str.substr(0, cutAt));
    }
    str = str.substr(cutAt + 1);
  }

  if (str.length() > 0) {
    results.push_back(str);
  }

  return results;
}

/*
************************************************************************
*+
* FUNCTION: stringToInt
*
* RETURNS: A pass/fail bool
*
* DESCRIPTION: Takes a string and converts it to an int. Pass-by-reference. 
*
* [NOTES:]: I found this algorithm on an internet message board. 
*
*-
************************************************************************
*/
bool stringToInt(const string &s, int &i) {
  istringstream myStream(s);

  if (myStream >> i)
    return true;
  else
    return false;
}

/*
************************************************************************
*+
* FUNCTION: stringToFloat
*
* RETURNS: A pass/fail bool
*
* DESCRIPTION: Takes a string and converts it to a float. Pass-by-reference. 
*
* [NOTES:]: I found this algorithm on an internet message board.
*
*-
************************************************************************
*/
bool stringToFloat(const string &s, float &i) {
  istringstream myStream(s);

  if (myStream >> i)
    return true;
  else
    return false;
}

/**
 * ------------------- Helper Class Function Definitions -----------------------
 * (Maybe this should be in its own file, but these functions haven't caused
 *   any problems since they were created, and now they are out of the way)
 **/

/************************************************************************
 *+
 * CLASS: Slit
 *
 * DESCRIPTION: Defines a slit object to be used in the slit selection
 * algorithm.
 *
 *-
 ************************************************************************
 */
Slit::Slit() {
  this->id = 0;
  this->priority = ' ';
  this->slitStart = this->slitEnd = this->slitLength = -1;
}

Slit::Slit(int Id, char Priority) {
  this->id = Id;
  this->priority = Priority;
  this->slitStart = this->slitEnd = this->slitLength = -1;
}

Slit::Slit(int Id, char Priority, float start, float end) {
  this->id = Id;
  this->priority = Priority;
  this->slitStart = start;
  this->slitEnd = end;
  this->slitLength = end - start;
}

Slit::Slit(int Id, char Priority, float start, float end, float len, float posL,
	   float posW, float width, float top, float bottom, float specPL,
	   float specPW, float Angle, float wiggleFact, string ln) {
  this->id = Id;
  this->priority = Priority;
  this->slitStart = start;
  this->slitEnd = end;
  this->slitLength = len;
  this->ccdL = posL;
  this->ccdW = posW;
  this->slitWidth = width;
  this->line = ln;
  this->slitTop = top;
  this->slitBottom = bottom;
  this->specPosL = specPL;
  this->specPosW = specPW;
  this->angle = Angle;

  if (Priority != '0' && wiggleFact > 0) {
    // Wiggle on!
    this->wiggleRoom = wiggleFact * slitLength;
    this->maxWiggle = this->wiggleRoom;
    this->wiggleUsed = false;
  } else {
    // Wiggle off or acq object.
    this->wiggleRoom = 0;
    this->maxWiggle = 0;
    this->wiggleUsed = true;
  }
}

/* FUNCTION changePosition
 *  Change the position of this slit.
 *
 */
int Slit::changePosition(float deltaW, float deltaL, float pixelScale,
			 string instType) {
  // Edit slit class values.
  this->ccdL += deltaL;
  this->slitStart += deltaL;
  this->slitEnd += deltaL;
  this->ccdW += deltaW;
  this->slitTop += deltaW;
  this->slitEnd += deltaW;

  vector < string > lineData;
  float slitPosW = 0.0;
  float slitPosL = 0.0;

  // For conversion issues, bit of a hack, but it works :( .
  string one, two, three, four, five, six, seven, ten, eleven, twelve,
    thirteen, fourteen, fifteen, sixteen;

  // Unpack, edit and repack 'line' string.

  lineData = stringSplit(this->line, "\t");

  one = lineData[0];
  two = lineData[1];
  three = lineData[2];
  four = lineData[3];
  five = lineData[4];
  six = lineData[5];
  seven = lineData[6];
  ten = lineData[9];
  eleven = lineData[10];
  twelve = lineData[11];
  thirteen = lineData[12];
  fourteen = lineData[13];
  fifteen = lineData[14];

  // Wave_ccd junk.
  sixteen = "";
  if (lineData.size() > 15) {
    sixteen = lineData[15];

    // Remove trailing newline and whitepace.
    sixteen.erase(sixteen.find_last_not_of(" \t\r\n") + 1);
  } else {
    fifteen.erase(fifteen.find_last_not_of(" \t\r\n") + 1);
  }

  if (instType == "F2") {
    stringToFloat(lineData[7], slitPosL);
    stringToFloat(lineData[8], slitPosW);
  } else if (instType == "GMOS" || instType == "GMOS-S"
	     || instType == "GMOS-N") {

    stringToFloat(lineData[7], slitPosW);
    stringToFloat(lineData[8], slitPosL);
  }

  slitPosL += deltaL * pixelScale;
  slitPosW += deltaW * pixelScale;

  char lineChar[256];
  if (instType == "F2") {
    sprintf(lineChar,
	    "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%8.6f\t%8.6f\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n",
	    one.c_str(), two.c_str(), three.c_str(), four.c_str(),
	    five.c_str(), six.c_str(), seven.c_str(), slitPosL, slitPosW,
	    ten.c_str(), eleven.c_str(), twelve.c_str(), thirteen.c_str(),
	    fourteen.c_str(), fifteen.c_str(), sixteen.c_str());
  } 
  else if (instType == "GMOS" || instType == "GMOS-S" || instType == "GMOS-N") {
    sprintf(lineChar,
	    "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%8.6f\t%8.6f\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n",
	    one.c_str(), two.c_str(), three.c_str(), four.c_str(),
	    five.c_str(), six.c_str(), seven.c_str(), slitPosW, slitPosL,
	    ten.c_str(), eleven.c_str(), twelve.c_str(), thirteen.c_str(),
	    fourteen.c_str(), fifteen.c_str(), sixteen.c_str());
  }

  this->line = lineChar;

  return 0;
}

bool Slit::stringToFloat(const string &s, float &i) {
  istringstream myStream(s);

  if (myStream >> i)
    return true;
  else
    return false;
}

vector<string> Slit::stringSplit(string str, string delim) {
  vector < string > results;

  int cutAt;
  while (int((cutAt = str.find_first_of(delim))) != int(str.npos)) {
    if (cutAt > 0) {
      results.push_back(str.substr(0, cutAt));
    }
    str = str.substr(cutAt + 1);
  }

  if (str.length() > 0) {
    results.push_back(str);
  }

  return results;
}

/*
************************************************************************
*+
* CLASS: GraphNode
*
* DESCRIPTION: Element class for following Graph class.
*
*-
************************************************************************
*/

GraphNode::GraphNode() {
  this->id = -1;
}

GraphNode::GraphNode(int Id) {
  this->id = Id;
}

// Add an item to this GraphNode's adjList.
void GraphNode::addAdj(int addID) {
  this->adjList.push_back(addID);
}

// Remove an id from this GraphNode's adjList.
void GraphNode::removeAdj(int targetID) {
  vector<int>::iterator it;

  // search for target item and remove:
  for (it = this->adjList.begin(); it != this->adjList.end(); it++) {
    if (*it == targetID) {
      this->adjList.erase(it);
      break;
    }
  }
}

vector<int> GraphNode::getList() {
  return this->adjList;
}

int GraphNode::getDegree() {
  return this->adjList.size();
}

/*
************************************************************************
*+
* CLASS: Graph
*
* DESCRIPTION: An adjacency-list type graph data structure.
*
*
*-
************************************************************************
*/

// Adds a GraphNode to the Graph
void Graph::addNode(int targetID) {
  elements[targetID] = GraphNode(targetID);
}

// Removes a GraphNode from the Graph
void Graph::removeNode(int nid) {
  vector<GraphNode>::iterator it;

  // Remove all edges connected to this node.
  this->clearEdges(&elements[nid]);

  // Remove element.
  elements.erase(nid);
}

// Clears the graph.
void Graph::clear() {
  map<int, GraphNode>::iterator it;

  for (it = elements.begin(); it != elements.end(); it++) {
    removeNode((*it).first);
  }
}

// Adds an edge between two GraphNodes.
void Graph::addEdge(GraphNode * node1, GraphNode * node2) {
  node1->addAdj(node2->id);
  node2->addAdj(node1->id);
}

// Removes an edge between two GraphNodes.
void Graph::removeEdge(GraphNode * node1, GraphNode * node2) {
  node1->removeAdj(node2->id);
  node2->removeAdj(node1->id);
}

// Fetches a reference to a GraphNode.
GraphNode * Graph::getNode(int nid) {
  return &elements[nid];
}

// Clears all edges connected to to given node.
void Graph::clearEdges(GraphNode * node) {
  vector<int>::iterator it;
  vector<int> list = node->getList();

  for (it = list.begin(); it != list.end(); it++) {
    this->removeEdge(node, getNode((*it)));
  }
}

// Returns true if the Graph is empty of GraphNodes.
bool Graph::empty() {
  return elements.size() == 0;
}

int Graph::size() {
  return elements.size();
}

// Returns a list of nodes that have the lowest degrees.
vector<int> Graph::getMinDegree() {
  // Map iterator.
  map<int, GraphNode>::iterator it;

  // Map container for nodes with min cardinality.
  vector<int> minNodeIds;
  int min = 32000;
  int curID;
  int curDegree;

  // Find minimum degree and populate min degree list.
  for (it = elements.begin(); it != elements.end(); it++) {
    curDegree = (*it).second.getDegree();
    if (curDegree < min) {
      // New lowest degree.
      // Erase low degree list and add this item.
      min = curDegree;
      curID = (*it).first;
      minNodeIds.clear();
      minNodeIds.push_back(curID);
    } 
    else if (curDegree == min) {
      // Another item with lowest degree. Add to map.

      curID = (*it).first;
      minNodeIds.push_back(curID);
    }
  }

  return minNodeIds;
}
